// DO NOT EDIT THIS FILE. IT WAS AUTOMATICALLY GENERATED BY:
//
//   ucd-generate general-category ucd-15.0.0 --exclude unassigned --enum --fst-dir benches/tables/fst
//
// Unicode version: 15.0.0.
//
// ucd-generate 0.2.15 is available on crates.io.

pub const GENERAL_CATEGORY_ENUM: &'static [&'static str] = &[
    "Close_Punctuation",
    "Connector_Punctuation",
    "Control",
    "Currency_Symbol",
    "Dash_Punctuation",
    "Decimal_Number",
    "Enclosing_Mark",
    "Final_Punctuation",
    "Format",
    "Initial_Punctuation",
    "Letter_Number",
    "Line_Separator",
    "Lowercase_Letter",
    "Math_Symbol",
    "Modifier_Letter",
    "Modifier_Symbol",
    "Nonspacing_Mark",
    "Open_Punctuation",
    "Other_Letter",
    "Other_Number",
    "Other_Punctuation",
    "Other_Symbol",
    "Paragraph_Separator",
    "Private_Use",
    "Space_Separator",
    "Spacing_Mark",
    "Surrogate",
    "Titlecase_Letter",
    "Uppercase_Letter",
];

pub static GENERAL_CATEGORY: ::once_cell::sync::Lazy<
    ::fst::Map<&'static [u8]>,
> = ::once_cell::sync::Lazy::new(|| {
    ::fst::Map::from(
        ::fst::raw::Fst::new(&include_bytes!("general_category.fst")[..])
            .unwrap(),
    )
});
