// DO NOT EDIT THIS FILE. IT WAS AUTOMATICALLY GENERATED BY:
//
//   ucd-generate names ucd-15.0.0 --no-aliases --no-hangul --no-ideograph --fst-dir benches/tables/fst
//
// Unicode version: 15.0.0.
//
// ucd-generate 0.2.15 is available on crates.io.

pub static NAMES: ::once_cell::sync::Lazy<::fst::Map<&'static [u8]>> =
    ::once_cell::sync::Lazy::new(|| {
        ::fst::Map::from(
            ::fst::raw::Fst::new(&include_bytes!("names.fst")[..]).unwrap(),
        )
    });
